"""
Genetic Algorithm

but without first generation inbreeding
"""

import torch
import einx
from einx import get_at, rearrange

# constants

GOAL = 'Attention is all you need'

POP_SIZE = 100
MUTATION_RATE = 0.04
FRAC_FITTEST_SURVIVE = 0.25
FRAC_TOURNAMENT = 0.25
ELITE_FRAC = 0.05

# encode and decode functions

def encode(s):
    return torch.tensor([ord(c) for c in s])

def decode(t):
    return ''.join([chr(i) for i in t.tolist()])

# derived constants

gene_length = len(GOAL)
gene_midpoint = gene_length // 2
target_gene = encode(GOAL)

keep_fittest_len = int(POP_SIZE * FRAC_FITTEST_SURVIVE)
num_elite = int(ELITE_FRAC * POP_SIZE)
num_repro_and_mutate = keep_fittest_len - num_elite
num_tournament_contenders = int(num_repro_and_mutate * FRAC_TOURNAMENT)
num_children = POP_SIZE - keep_fittest_len
num_mutate = MUTATION_RATE * gene_length

assert num_tournament_contenders >= 2

# genetic algorithm

generation = 1

parent_ids = torch.full((POP_SIZE, 2), -1, dtype = torch.long)
pool = torch.randint(0, 255, (POP_SIZE, gene_length))

while True:
    print(f"\n\ngeneration {generation}\n")

    # sort population by fitness

    fitnesses = 1. / torch.square(pool - target_gene).sum(dim = -1)

    indices = fitnesses.sort(descending = True).indices
    pool, parent_ids, fitnesses = pool[indices], parent_ids[indices], fitnesses[indices]

    # keep the fittest

    pool, parent_ids, fitnesses = pool[:keep_fittest_len], parent_ids[:keep_fittest_len], fitnesses[:keep_fittest_len]

    # display every generation

    for gene, fitness in zip(pool, fitnesses):
        print(f"{decode(gene)} ({fitness.item():.3f})")

    # solved if any fitness is inf

    if (fitnesses == float('inf')).any():
        break

    # elites can pass directly to next generation

    elites, pool = pool[:num_elite], pool[num_elite:]
    elites_fitnesses, fitnesses = fitnesses[:num_elite], fitnesses[num_elite:]
    elites_parent_ids, parent_ids = parent_ids[:num_elite], parent_ids[num_elite:]

    elites_parent_ids.fill_(-1)

    # deterministic tournament selection
    # 2 tournaments - the second tournament removes all contestants with shared parents with 1st winner

    first_contender_ids = torch.randn((num_children, num_repro_and_mutate)).argsort(dim = -1)[..., :num_tournament_contenders]
    first_participants, participants_parent_ids, tournaments = pool[first_contender_ids], parent_ids[first_contender_ids], fitnesses[first_contender_ids]

    first_winner = tournaments.topk(1, dim = -1, largest = True, sorted = False).indices
    first_winner = rearrange('p 1 -> p', first_winner)

    first_parent_ids = get_at('p [t] i, p -> p i', participants_parent_ids, first_winner)

    # second tournament, masking out any siblings to first winners

    contender_scores = torch.randn((num_children, num_repro_and_mutate))
    self_mask = rearrange('i -> i 1', first_winner) == torch.arange(num_repro_and_mutate)
    contender_scores = torch.where(self_mask, 1e6, contender_scores)

    sibling_mask = (rearrange('p i -> p 1 i 1', first_parent_ids) == rearrange('c j -> 1 c 1 j', parent_ids))
    valid_parent_mask = (rearrange('p i -> p 1 i 1', first_parent_ids) != -1) & (rearrange('c j -> 1 c 1 j', parent_ids) != -1)
    num_shared_parents = (sibling_mask & valid_parent_mask).float().sum(dim = (-1, -2))
    contender_scores += num_shared_parents * 1e3

    second_contender_ids = contender_scores.argsort(dim = -1)[..., :num_tournament_contenders]
    second_participants, second_tournaments = pool[second_contender_ids], fitnesses[second_contender_ids]
    second_winner = second_tournaments.topk(1, dim = -1, largest = True, sorted = False).indices
    second_winner = rearrange('p 1 -> p', second_winner)

    # get parents

    first_ids = get_at('p [t], p -> p', first_contender_ids, first_winner)
    second_ids = get_at('p [t], p -> p', second_contender_ids, second_winner)

    new_parent_ids = torch.stack((first_ids, second_ids), dim = -1)

    parent1 = get_at('p [t] g, p -> p g', first_participants, first_winner)
    parent2 = get_at('p [t] g, p -> p g', second_participants, second_winner)

    # cross over recombination of parents

    children = torch.cat((parent1[:, :gene_midpoint], parent2[:, gene_midpoint:]), dim = -1)

    pool = torch.cat((pool, children))

    parent_ids.fill_(-1)
    parent_ids = torch.cat((parent_ids, new_parent_ids))

    # mutate genes in population

    mutate_mask = torch.randn(pool.shape).argsort(dim = -1) < num_mutate
    noise = torch.randint(0, 2, pool.shape) * 2 - 1
    pool = torch.where(mutate_mask, pool + noise, pool)
    pool.clamp_(0, 255)

    # add back the elites

    pool = torch.cat((elites, pool))
    parent_ids = torch.cat((elites_parent_ids, parent_ids))

    generation += 1
